#
# Implementation of support for simulation with uartlite in ModelSim.
#
# Copyright (c) 2008-2011 Xilinx, Inc.  All rights reserved. 
#
# Usage: In modelsim:
#
#   When starting simulation:
#     VSIM> source sim_uart.tcl
#     VSIM> uart_start
#
#   After restarting simulation:
#     VSIM> restart -f
#     VSIM> uart_restart
#
# stefana  10/07/08  Improved to support any path name to the uartlite instance
#                    Don't fail if terminal window is already open
# stefana  11/07/04  Corrected SendRX to filter empty input from the terminal
#
# $Id$
#

# Procedure that is called within Modelsim 
proc uart_start {} {
  # path variables
  global tx
  global rx

  # path to UART TX and UART RX (assume there is only one for now)
  set tx_instance [find instances -recursive opb_uartlite_tx_i]
  set rx_instance [find instances -recursive opb_uartlite_rx_i]
  if {[string length $tx_instance] == 0 || [string length $rx_instance] == 0} {
    set tx_instance [find instances -recursive uartlite_tx_i]
    set rx_instance [find instances -recursive uartlite_rx_i]
  }
  if {[string length $tx_instance] == 0 || [string length $rx_instance] == 0} {
    echo "** ERROR: uart_start: could not find path to UART core"
  } else {
    # define terminal
    destroy .term
    toplevel .term
    wm title .term "ModelSim - Terminal window for uartlite RS-232 serial interface"
    text .term.t -width 80 -height 50
    pack .term.t -fill both -expand 1

    set tx [lindex [split [lindex $tx_instance 0]] 0]

    # override UART TX VHDL signals statically, 
    # never full, always empty, no TX data
    force -freeze $tx/TX_Buffer_Full 0
    force -freeze $tx/TX_Buffer_Empty 1
    force -freeze $tx/TX 0

    # grab character every time new byte writen to UART TX
    when  "$tx/write_tx_fifo'event and $tx/write_tx_fifo='1'" {
      # Also handle verilog syntax, e.g. "8'd45"
      set tx_data [examine -decimal $tx/tx_data]
      if {[regexp {8'd([0-9]*)} $tx_data match num]} {
        set tx_data $num
      }

      set char [format %c $tx_data]
      # echo "TX $::now $char \"$tx\""
      if {$char ne "\r"} {
        .term.t insert end $char
      }
      .term.t see end
    }

    # path to UART RX
    set rx [lindex [split [lindex $rx_instance 0]] 0]

    # override UART RX VHDL signals statically
    # don't model any error
    force -freeze $rx/RX_Frame_Error 0
    force -freeze $rx/RX_Parity_Error 0

    # initial values on UART RX
    force -freeze $rx/fifo_wr 0
    force -freeze $rx/fifo_din 000000000

    # for UART RX bind tcl SendRX to terminal input
    entry .term.e
    pack .term.e -side bottom -fill x
    bind .term.e <Key> "SendRX %A"
  }
}

# when modelsim restarts do uart_restart as well
proc uart_restart {} {
  nowhen *
  destroy .term
  uart_start
}

# Set which character to send then 
# schedule wait for rising clock to write character to RX FIFO
proc SendRX {RX_Char} {
  global rx
  global RX_Force
  # echo "SendRX $::now $RX_Char \"$rx\""
  if {[scan $RX_Char %c RX_IntChar] != -1} {
    set RX_Force 10\#$RX_IntChar
    # echo "SendRX: $::now $RX_Force"
    if {$RX_Char eq "\n" || $RX_Char eq "\r"} {
      .term.e delete 0 end
    }
    when -label SetEdge "$rx/clk'event and $rx/clk='1'" {   
      SetRX
    }
  }
}

# Rising clock, write character to RX FIFO then
# schedule no write at next clock edge
proc SetRX {} {
  global rx
  global RX_Force
  # echo "SetRX $::now"
  nowhen SetEdge;
  force -freeze $rx/fifo_wr 1
  force -freeze $rx/fifo_din(1 to 8) $RX_Force
  when -label ClearEdge "$rx/clk'event and $rx/clk='1'" {
    ClearRX
  }
}

# Rising clock, write enable to RX FIFO deasserted
proc ClearRX {} {
  global rx
  # echo "ClearRX $::now"
  nowhen ClearEdge
  force -freeze $rx/fifo_wr 0
}

