/*
 * Example of using nested fast interrupts in MicroBlaze.
 *
 * The lower priority interrupt handler prints "1" at the start
 * and "2" at the end of execution. The highest priority handler
 * prints "0" when it is executed. Since the higher priority
 * interrupt is nested, and preempts the lower priority handler,
 * the output is "102" repeated MAX_COUNT times.
 *
 * The AXI Interrupt Controller must be set up to Enable Fast
 * Interrupt Logic and to Enable Interrupt Level Register to be
 * able to use nested fast interrupts.
 *
 * With normal interrupts the intc driver handles nested
 * interrupts in XIntc_DeviceInterruptHandler, including saving
 * and restoring the interrupt return address and Interrupt Level
 * Register, and there is no need to do anything special in user
 * code to utilize nested interrupts.
 */

#include <stdio.h>
#include <xintc.h>
#include <xtmrctr.h>

#define MAX_COUNT 5
#define DELAY_MAX 1000
#define RESET_VALUE 0xFFFF0000

static volatile int count;

/*
 * Interrupt controller base address used for saving and restoring
 * the Interrupt Level Register in fast interrupt handlers.
 */
static UINTPTR Intc_BaseAddress;

/*
 * Timer counter base addresses used for acknowledging interrupts.
 */
static UINTPTR TmrCtr_BaseAddress[2];

void TmrCtr_Acknowledge(UINTPTR BaseAddress);

/*
 * Declarations of fast interrupt handlers with attributes.
 */
void tmrctr_0_handler() __attribute ((fast_interrupt));
void tmrctr_1_handler() __attribute ((fast_interrupt));

/* Highest priority fast interrupt handler.
 *
 * This handler will interrupt the lower priority handler,
 * to demonstrate the nested interrupt capability.
 *
 * Since this is the highest priority interrupt, it can execute
 * with interrupts disabled, and does not need to manage the
 * AXI Interrupt Controller Interrupt Level Register.
 */
void tmrctr_0_handler()
{
	/*
	 * Note that it is normally not a good idea to use print in interrupt
	 * handlers. We just do it to show what is going on.
	 */
    print("0");

    /*
	 * The actual code to handle the interrupt. Here we just increment
	 * the interrupt count.
     */
    count++;

    /*
     * Acknowledge the interrupt.
     */
    TmrCtr_Acknowledge(TmrCtr_BaseAddress[0]);
}

/* Lower priority fast interrupt handler.
 *
 * This handler will be interrupted by the highest priority handler,
 * to demonstrate the nested interrupt capability.
 *
 * The handler must save the return address and the AXI Interrupt
 * Controller Interrupt Level Register (ILR), set ILR to the current
 * interrupt level, and enable interrupts before handling the interrupt.
 *
 * Finally it must restore the previous Interrupt Level Register value
 * and return address with interrupts disabled.
 */
void tmrctr_1_handler()
{
	volatile u32 R14, ILR;
	volatile int delay = 0;

	/*
	 * Save interrupt return address from register r14, save
	 * Interrupt Level Register, and set it to current interrupt level.
	 */
    R14 = mfgpr(r14);
	ILR = Xil_In32(Intc_BaseAddress + XIN_ILR_OFFSET);
	Xil_Out32(Intc_BaseAddress + XIN_ILR_OFFSET, 1);

	/*
     * Acknowledge the interrupt, and enable nested interrupts. Note
     * that the interrupt must be acknowledged before enabling interrupts.
     */
    TmrCtr_Acknowledge(TmrCtr_BaseAddress[1]);
	Xil_ExceptionEnable();

	/*
	 * The actual code to handle the interrupt. Here we just spend some
	 * time in the handler to allow the nested highest priority interrupt
	 * to occur.
	 *
	 * Note that it is normally not a good idea to use print in interrupt
	 * handlers. We just do it to show what is going on.
	 */
	print("1");
	while (delay++ < DELAY_MAX) {};
	print("2");

    /*
	 * Restore ILR and interrupt return address to register r14
	 * with interrupts disabled.
	 */
	Xil_ExceptionDisable();
	Xil_Out32(Intc_BaseAddress + XIN_ILR_OFFSET, ILR);
	mtgpr(r14, R14);
}

int IntcInterruptSetup(XIntc *IntcInstancePtr, u16 DeviceId)
{

	int Status;

	/*
	 * Initialize the interrupt controller driver so that it is
	 * ready to use.
	 */
	Status = XIntc_Initialize(IntcInstancePtr, DeviceId);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}

	/*
	 * Save the interrupt controller base address for use in fast
	 * interrupt handlers.
	 */
	Intc_BaseAddress = IntcInstancePtr->BaseAddress;

	/*
	 * Initialize the exception table.
	 */
	Xil_ExceptionInit();

	/*
	 * Register the interrupt controller handler with the exception table.
	 */
	Xil_ExceptionRegisterHandler(XIL_EXCEPTION_ID_INT,
			(Xil_ExceptionHandler)XIntc_DeviceInterruptHandler,
			(void*) 0);

	/*
	 * Enable exceptions.
	 */
	Xil_ExceptionEnable();

	/*
	 * Start the interrupt controller such that interrupts are enabled for
	 * all devices that cause interrupts.
	 */
	Status = XIntc_Start(IntcInstancePtr, XIN_REAL_MODE);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}

	return XST_SUCCESS;
}

int TmrCtrSetup(XIntc* IntcInstancePtr,
                XTmrCtr* TmrCtrInstancePtr,
				XFastInterruptHandler TimerCounterHandler,
                u16 DeviceId,
                u16 IntrId)
{
	int Status;

	/*
	 * Initialize the timer counter so that it's ready to use,
	 * specify the device ID that is generated in xparameters.h
	 */
	Status = XTmrCtr_Initialize(TmrCtrInstancePtr, DeviceId);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}

	/*
	 * Save the timer counter base address for use in fast interrupt handlers.
	 */
	TmrCtr_BaseAddress[DeviceId] = TmrCtrInstancePtr->BaseAddress;

	/*
	 * Connect a device driver handler that will be called when an interrupt
	 * for the device occurs, the device driver handler performs the specific
	 * interrupt processing for the device
	 */
	Status = XIntc_ConnectFastHandler(IntcInstancePtr, IntrId,
				(XFastInterruptHandler)TimerCounterHandler);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}

	/*
	 * Enable the interrupt for the timer counter
	 */
	XIntc_Enable(IntcInstancePtr, IntrId);

	/*
	 * Enable the interrupt of the timer counter so interrupts will occur
	 * and use auto reload mode such that the timer counter will reload
	 * itself automatically and continue repeatedly, without this option
	 * it would expire once only.
	 */
	XTmrCtr_SetOptions(TmrCtrInstancePtr, 0,
                XTC_INT_MODE_OPTION | XTC_AUTO_RELOAD_OPTION);

	/*
	 * Set a reset value for the timer counter such that it will expire
	 * earlier than letting it roll over from 0, the reset value is loaded
	 * into the timer counter when it is started.
	 */
	XTmrCtr_SetResetValue(TmrCtrInstancePtr, 0, RESET_VALUE);

	/*
	 * Start the timer counter such that it's incrementing by default,
	 * then wait for it to timeout a number of times.
	 */
	XTmrCtr_Start(TmrCtrInstancePtr, 0);

	return XST_SUCCESS;
}

void TmrCtr_Acknowledge(UINTPTR BaseAddress)
{
	/*
	 * Acknowledge the interrupt by clearing the interrupt bit in the
	 * timer control status register. The interrupt is cleared by writing
	 * a 1 to the interrupt bit of the register without changing any of
	 * the other bits.
	 */
	u32 TCSR = XTmrCtr_ReadReg(BaseAddress, 0, XTC_TCSR_OFFSET);
	XTmrCtr_WriteReg(BaseAddress, 0, XTC_TCSR_OFFSET,
			TCSR | XTC_CSR_INT_OCCURED_MASK);
}

int main()
{
    XTmrCtr TmrCtr;
    XIntc Intc;
    int Status;

    count = 0;

    print("---Entering main---\r\n");

    /* Set up interrupt controller */
    Status = IntcInterruptSetup(&Intc, XPAR_AXI_INTC_0_DEVICE_ID);
    if (Status != XST_SUCCESS)
        return XST_FAILURE;

    /* Set up timers */
    Status = TmrCtrSetup(&Intc, &TmrCtr, tmrctr_1_handler,
    		XPAR_TMRCTR_1_DEVICE_ID, XPAR_INTC_0_TMRCTR_1_VEC_ID);
    if (Status != XST_SUCCESS)
        return XST_FAILURE;

    Status = TmrCtrSetup(&Intc, &TmrCtr, tmrctr_0_handler,
    		XPAR_TMRCTR_0_DEVICE_ID, XPAR_INTC_0_TMRCTR_0_VEC_ID);
    if (Status != XST_SUCCESS)
        return XST_FAILURE;

    /* Wait until interrupt count reaches maximum */
    while (count < MAX_COUNT) {};

    /* Disable interrupts and stop timers */
	Xil_ExceptionDisable();

    XTmrCtr_Stop(&TmrCtr, XPAR_TMRCTR_0_DEVICE_ID);
    XTmrCtr_Stop(&TmrCtr, XPAR_TMRCTR_1_DEVICE_ID);

    print("\r\n---Exiting main---\r\n");

    return 0;
}
